#!/usr/bin/env -S guile -e main -s
coding: utf-8
!#

(use-modules
 (ice-9 exceptions)
 (ice-9 match)
 (ice-9 format)
 (ice-9 peg)
 (ice-9 rdelim)
 (ice-9 regex)
 (juniority database)
 (juniority grammar)
 (juniority util)
 (srfi srfi-1)
 (srfi srfi-43)
 )

(define (main args)
  (match args
    ((bin "help"   . rest) (display-help bin))
    ((bin "-h"     . rest) (display-help bin))
    ((bin "--help" . rest) (display-help bin))
    ((bin db-path) (main-import db-path))
    ((bin . _) (display-help bin))
    ))
(define (display-help bin)
  (format #t "Import:
In order to import your existing records, it must be in the following format:
    stmt        = volume-stmt | idol-stmt | tag-stmt
    volume-stmt = studio-[0-9]+ idol-assoc* tag-assoc* pub?
    idol-stmt   = idol-tag tag-assoc* dob?

    idol-assoc  = idol-tag | idol-untag
    idol-tag    = @'name'
    idol-untag  = !@'name'

    tag-assoc   = tag | untag
    tag         = 'name'
    untag       = !'name'

    dob         = date
    pub         = date
    date        = nnnn-nn-nn
    n           = [0-9]
    studio      = [A-Z]+

Example:
    @'高橋ひかる' 'Takahashi' 'Hikaru' 2001-09-22
    BSTAR-001 @'高橋ひかる'
    BSTAR-019 @'高橋ひかる'
    @'蒼井玲奈' 'Aoi' 'Reina' 1996-03-19
    ICDV-30223 @'蒼井玲奈' 'thong' 'butt' 'catgirl' 'lollipop' 'popsicle'
    ...

Notes:
- Idol statements must precede volume statements that reference them.
- Duplicate idol and volume statements are treated as updates.
- (Idol-)Untag expressions remove a previously-associated tag (or idol).

Feed your records to juniority to bootstrap your database:

    ~a import path/to/db.sqlite3 < path/to/import.jun

Use import mode in the future to add records to your database. For manual entry, combine it with a heredoc for convenience:

    ~a import path/to/db.sqlite3 << EOF
    ...
    EOF
" bin bin))

(define (main-import db-path)
  (let ((db (open-db db-path))
        )
    (migrate-db db)
    (do ((line (read-line) (read-line)))
        ((eof-object? line))
      (process-import-line db (peg:tree (match-pattern import-line line)))
      )))

(define (process-import-line db line)
  (let ()
    (match line
      (('volume-entry ('volume identifier) 'idols 'tags ('date pub-date))
       (import-volume db identifier pub-date (list) (list)))
      (('volume-entry ('volume identifier) idols 'tags ('date pub-date))
       (import-volume db identifier pub-date (cdr idols) (list)))
      (('volume-entry ('volume identifier) 'idols tags ('date pub-date))
       (import-volume db identifier pub-date (list) (cdr tags)))
      (('volume-entry ('volume identifier) idols tags ('date pub-date))
       (import-volume db identifier pub-date (cdr idols) (cdr tags)))
      (('volume-entry ('volume identifier) 'idols 'tags)
       (import-volume db identifier #nil (list) (list)))
      (('volume-entry ('volume identifier) idols 'tags)
       (import-volume db identifier #nil (cdr idols) (list)))
      (('volume-entry ('volume identifier) 'idols tags)
       (import-volume db identifier #nil (list) (cdr tags)))
      (('volume-entry ('volume identifier) idols tags)
       (import-volume db identifier #nil (cdr idols) (cdr tags)))

      (('idol-entry ('idol-tag jp-name) 'tags ('date birth-date))
       (import-idol db jp-name birth-date (list)))
      (('idol-entry ('idol-tag jp-name) tags ('date birth-date))
       (import-idol db jp-name birth-date (cdr tags)))
      (('idol-entry ('idol-tag jp-name) 'tags)
       (import-idol db jp-name #nil (list)))
      (('idol-entry ('idol-tag jp-name) tags)
       (import-idol db jp-name #nil (cdr tags)))
      (#f (noop))
    )))

(define (import-volume db identifier pub-date idols tags)
  (let* ((studio-callsign
          (first (string-split identifier (lambda (ch) (eq? #\- ch)))))

         (volume-id (param-query-one db select-volume-stmt identifier))
         (studio-id (param-query-one db select-studio-stmt studio-callsign))
         (idol-ids #nil)
         (tag-ids  #nil)

         (existing-volume? #f)
         )
    (if (vector? volume-id)
        (begin
          (set! volume-id (vector-ref volume-id 0))
          (set! existing-volume? #t)
          (log "import-volume: found volume ~a" volume-id))
        (begin
          (set!
           volume-id
           (vector-ref
            (param-query-one db insert-volume-stmt identifier pub-date)
            0))
          (log "import-volume: inserted volume ~a" volume-id)))
    (if (vector? studio-id)
        (begin
          (set! studio-id (vector-ref studio-id 0))
          (log "import-volume: found studio ~a" studio-id))
        (begin
          (set!
           studio-id
           (vector-ref
            (param-query-one db insert-studio-stmt studio-callsign)
            0))
          (log "import-volume: inserted studio ~a" studio-id)))
    (set!
     idol-ids
     (map
      (lambda (idol)
        (let* ((idol-name (second idol))
               (idol-result (param-query-one db select-idol-stmt idol-name))
               (idol-id
                (if (vector? idol-result)
                    (begin
                      (log "import-volume: found idol ~a" idol-name)
                      (vector-ref idol-result 0))
                    (raise-exception
                     (make-exception-with-message
                      (format #f "import-volume: idol not found ~a" idol-name))))))
          (list (first idol) idol-id)))
      idols))
    (set!
     tag-ids
     (map
      (lambda (tag)
        (let* ((tag-name (second tag))
               (tag-result (param-query-one db select-tag-stmt tag-name))
               (tag-id (if (vector? tag-result)
                           (begin
                             (log "import-volume: found tag ~a" tag-name)
                             (vector-ref tag-result 0))
                           (begin
                             (log "import-volume: inserted tag ~a" tag-name)
                             (vector-ref
                              (param-query-one db insert-tag-stmt tag-name)
                              0)))))
          (list (first tag) tag-id)))
      tags))

    (if existing-volume?
        (when pub-date
          (param-query-one db update-volume-pub-stmt pub-date volume-id)
          (log "import-volume: updated volume published date"))
        (begin
          (param-query-one db assoc-volume-studio-stmt volume-id studio-id)
          (log "import-volume: associated studio ~a" studio-id)))

    (let ((existing-tags
           (param-query-map db
                            (lambda (ct row) (vector-ref row 0))
                            select-volume-tag-ids-stmt volume-id))
          (existing-idols
           (param-query-map db
                            (lambda (ct row) (vector-ref row 0))
                            select-volume-idol-ids-stmt volume-id)))

      (for-each
       (lambda (idol)
         (match idol
           (('idol-tag idol-id)
            (unless (find (lambda (id) (eqv? idol-id id)) existing-idols)
              (param-query-one db assoc-volume-idol-stmt volume-id idol-id)
              (log "import-volume: associated idol ~a" idol-id)))
           (('idol-untag idol-id)
            (when (find (lambda (id) (eqv? idol-id id)) existing-idols)
              (param-query-one db deassoc-volume-idol-stmt volume-id idol-id)
              (log "import-volume: deassociated idol ~a" idol-id)))
           (other (log "import-volume: unmatched idol pattern: ~a" other))
           )
         )
       idol-ids)
      (for-each
       (lambda (tag)
         (match tag
           (('tag tag-id)
            (unless (find (lambda (id) (eqv? tag-id id)) existing-tags)
              (param-query-one db assoc-volume-tag-stmt volume-id tag-id)
              (log "import-volume: associated tag ~a" tag-id)))
           (('untag tag-id)
            (when (find (lambda (id) (eqv? tag-id id)) existing-tags)
              (param-query-one db deassoc-volume-tag-stmt volume-id tag-id)
              (log "import-volume: deassociated tag ~a" tag-id)))
           (other (log "import-volume: unmatched tag pattern: ~a" other))))
       tag-ids)
      )))

(define (import-idol db jp-name birth-date tags)
  (let* ((idol-id (param-query-one db select-idol-stmt jp-name))
         (tag-ids (list))

         (existing-idol? #f)
         )
    (if (vector? idol-id)
        (begin
          (set! idol-id (vector-ref idol-id 0))
          (set! existing-idol? #t)
          (log "import-idol: found idol ~a" idol-id))
        (begin
          (set!
           idol-id
           (vector-ref (param-query-one db insert-idol-stmt jp-name birth-date) 0))
          (log "import-idol: inserted idol ~a" idol-id)))
    (set!
     tag-ids
     (map
      (lambda (tag)
        (let* ((tag-name (second tag))
               (tag-result (param-query-one db select-tag-stmt tag-name))
               (tag-id
                (if (vector? tag-result)
                    (begin
                      (log "import-idol: found tag ~a" tag-name)
                      (vector-ref tag-result 0))
                    (begin
                      (log "import-idol: inserted tag ~a" tag-name)
                      (vector-ref
                       (param-query-one db insert-tag-stmt tag-name)
                       0)))))
          (list (first tag) tag-id)))
      tags))

    (when (and existing-idol? birth-date)
      (param-query-one db update-idol-dob-stmt birth-date idol-id)
      (log "import-idol: updated idol birth date"))

    (let ((existing-tags
           (param-query-map db
                            (lambda (ct row) (vector-ref row 0))
                            select-idol-tag-ids-stmt idol-id)))
      (for-each
       (lambda (tag)
         (match tag
           (('tag tag-id)
            (unless (find (lambda (id) (eqv? tag-id id)) existing-tags)
              (param-query-one db assoc-idol-tag-stmt idol-id tag-id)
              (log "import-idol: associated tag ~a" tag-id)))
           (('untag tag-id)
            (when (find (lambda (id) (eqv? tag-id id)) existing-tags)
              (param-query-one db deassoc-idol-tag-stmt idol-id tag-id)
              (log "import-idol: deassociated tag ~a" tag-id)))
           (other (log "import-idol: unmatched tag pattern: ~a" other))))
       tag-ids)
      )
    ))
